"""Launch a nextflow workflow."""

import argparse
import http.client as httplib
import json
import logging
from multiprocessing.connection import Listener
import os
import queue
import subprocess
import sys
import threading
from typing import Any, Dict, Union

from epi2melabs.workflows.database import get_session, Instance, Statuses
from epi2melabs.workflows.launcher import popen


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Execute a netflow workflow and update the database.",
        usage=(
            "invoke_nextflow -w epi2melabs/wf-alignment -i <instance_id> "
            "-w <workflow_name> -p <params_file> -r <revision> -m <meta_file> "
            "-wd <work_dir> -l <log_file> -s <stdout_file> -d <database> "
            "-rpc <port>"
        )
    )

    parser.add_argument(
        '-n',
        '--nextflow',
        required=True,
        default='nextflow',
        help='Path to the nextflow executable.'
    )

    parser.add_argument(
        '-i',
        '--id',
        required=True,
        help='ID of the database instance record to acquire and update.'
    )

    parser.add_argument(
        '-w',
        '--workflow',
        required=True,
        help='Path to or name of the workflow to be run.'
    )

    parser.add_argument(
        '-p',
        '--params',
        required=True,
        help='Path to the workflow params file.'
    )

    parser.add_argument(
        '-r',
        '--revision',
        required=False,
        default=None,
        help='Workflow revision to execute.'
    )

    parser.add_argument(
        '-wd',
        '--work_dir',
        required=True,
        help='Path to what should become the working directory.'
    )

    parser.add_argument(
        '-l',
        '--log_file',
        required=True,
        help='Path to which the logs should be written.'
    )

    parser.add_argument(
        '-s',
        '--std_out',
        required=True,
        help='Path to which the stdout should be written.'
    )

    parser.add_argument(
        '-m',
        '--meta_file',
        required=True,
        help=(
            'Path at which to write JSON file containing '
            'meta information such as the commands executed.'
        )
    )

    parser.add_argument(
        '-d',
        '--database',
        required=True,
        help='Path to the SQLITE database to update.'
    )

    parser.add_argument(
        '-rpc',
        '--rpc_port',
        required=True,
        help=(
            'Port by which to communicate via rpc. '
            'Overrides the "rpc" key within -m, if given.'
        )
    )

    parser.add_argument(
        '-wsl',
        action='store_true',
        help='Run command in wsl'
    )

    start = 0
    if 'invoke_nextflow' in sys.argv[0]:
        start = 1

    return parser.parse_args(sys.argv[start:])


def have_internet(use_shell=False):
    """Check if we have internet."""
    if use_shell:
        return not popen(
            ['ping', '-c 4', '8.8.8.8'],
            stdout=subprocess.DEVNULL).wait()
    # Thankyou https://stackoverflow.com/a/29854274
    conn = httplib.HTTPSConnection("8.8.8.8", timeout=5)
    try:
        conn.request("HEAD", "/")
        return True
    except Exception as e:
        logging.info(e)
        return False
    finally:
        conn.close()


#
# Functionality needed for implementing a sigint
# equivalent in windows
#
def _thread_process_wait(queue, proc):
    """Wait for a process to complete then notify main thread."""
    ret = proc.wait()
    queue.put(ret)


def _thread_ipc_listen(queue, port, _id):
    """Wait for a signal then notify main thread."""
    address = ('localhost', int(port))
    listener = Listener(address, authkey=bytes(
        _id, encoding='utf8'))
    conn = listener.accept()
    while True:
        msg = conn.recv()
        if msg[0] == 'close' and msg[1] == _id:
            break
    conn.close()
    listener.close()
    queue.put(False)


def kill_process_wsl(pidfile, stdout, stderr, threads=None):
    """End a process running in wsl."""
    popen(
        ['wsl', 'kill', f'$(cat {pidfile})'], windows=True,
        stdout=stdout, stderr=stderr).wait()
    if threads:
        [thread.join() for thread in threads]


def wait_for_workflow(proc, port, _id):
    """Wait for a process to end or an interrupt on windows."""
    q = queue.Queue()
    listener = threading.Thread(
        target=_thread_ipc_listen, args=(q, port, _id))
    waiter = threading.Thread(
        target=_thread_process_wait, args=(q, proc))
    listener.start()
    waiter.start()
    return q.get(), [listener, waiter]


def invoke(
    id: str, workflow: str, params: str, work_dir: str,
    log_file: str, meta_file: str, std_out: str, database: str,
    nextflow: str, rpc_port: str, revision: Union[str, None] = None,
    wsl: bool = False
) -> None:
    """Run nextflow workflow."""
    logging.basicConfig(
        format='invoke_nextflow <%(asctime)s>: %(message)s',
        level=logging.DEBUG)
    logging.info('Initialising.')

    # This somehow fixes when the path when it contains spaces
    nextflow = fr"{nextflow}"

    # Containers for the process and execution threads we'll use
    proc = None
    threads = []

    # Establish settings container
    meta: Dict[str, Any] = dict(
        id=id, workflow=workflow, params=params, work_dir=work_dir,
        log_file=log_file, std_out=std_out, database=database,
        nextflow=nextflow, revision=revision, rpc_port=rpc_port,
        wsl=wsl, wsl_pidfile=None, run_command=None, pull_command=None,
        offline=False)

    # Get command to run workflow
    run_command = [
        nextflow, '-log', log_file, 'run', workflow,
        '-params-file', params, '-w', work_dir,
        '-ansi-log', 'false']

    # Check if we can connect to the internet
    can_connect = have_internet(use_shell=not wsl)

    # Set up command to update workflow
    pull_command = None
    if can_connect:
        if not os.path.isfile(workflow):
            pull_command = [nextflow, 'pull', workflow]
    # Set offline mode if no connection
    else:
        logging.info(
            'Cannot connect to the internet,'
            'not updating workflow.')
        meta['offline'] = True
        run_command.append('-offline')

    # If we are using wsl, we need to adjust our command
    # to store the wsl process pid
    if wsl:
        logging.info('Setting command to run in WSL.')
        meta["wsl_pidfile"] = os.path.dirname(
            work_dir) + '/' + 'wsl_proc.pid'
        wsl_preexec = [
            'wsl', 'echo', '$$', '>', f'{meta["wsl_pidfile"]};', 'exec']
        run_command = wsl_preexec + run_command
        if pull_command:
            pull_command = ['wsl'] + pull_command

    # If we have specified a revision, add it to the
    # end of the command
    if revision:
        logging.info(f'Using revision {revision}.')
        run_command = run_command + ['-r', revision]

    # Get the invocation instance by id from the
    # database and update the pid with this script's
    # Note: this may become somewhat legacy after
    # switching to RPC for killing procs.
    db = get_session(database)
    invocation = db.query(Instance).get(id)
    pid = os.getpid()
    logging.info(f'The wrapper PID is {pid}.')
    invocation.pid = pid
    db.commit()

    # Get handles for outputs
    cli_logfile = open(std_out, 'a')
    stdout = cli_logfile
    stderr = cli_logfile

    try:
        # Update the workflow
        if pull_command:
            logging.info('Updating workflow.')
            meta["pull_command"] = " ".join(pull_command)
            logging.info(f'Command: {meta["pull_command"]}.')
            # We use our custom popen wrapper that helps to
            # prevent the creation of cmd windows on Windows,
            # and at the moment -wsl is synonymous with using
            # Windows.
            proc = popen(
                pull_command, windows=wsl, stdout=stdout,
                stderr=stderr)
            if proc.wait():
                logging.info(
                    'Could not update workflow, are you connected '
                    'to the internet?')

        # Invoke the workflow
        logging.info('Launching workflow.')
        meta["run_command"] = " ".join(run_command)
        logging.info(f'Command: {meta["run_command"]}.')
        proc = popen(
            run_command, windows=wsl, stdout=stdout,
            stderr=stderr)
        logging.info(f'The workflow PID is {proc.pid}.')

        # Set initial database status
        invocation.status = Statuses.LAUNCHED
        db.commit()

        # Dump the meta information to file
        with open(meta_file, 'w', encoding='utf8') as mf:
            mf.write(json.dumps(meta))

        # Listen for a sigint via rpc and/or wait
        # for the workflow to exit
        ret, threads = wait_for_workflow(
            proc, rpc_port, id)
        if ret is False:
            raise KeyboardInterrupt
        sys.exit(ret)

    # If we receive sigint, assume the process was
    # terminated intentionally and exit gracefully
    except KeyboardInterrupt:
        logging.info('Interrupt detected: terminating workflow.')
        if wsl and proc:
            kill_process_wsl(
                meta["wsl_pidfile"], stdout, stderr, threads)
        elif proc:
            proc.kill()
        invocation.status = Statuses.TERMINATED
        db.commit()
        sys.exit(0)

    except SystemExit as e:
        # If we receive system exit of 0, assume the process
        # ended peacefully and exit gracefully.
        if not e.code:
            logging.info('Workflow completed.')
            invocation.status = Statuses.COMPLETED_SUCCESSFULLY
            db.commit()
            sys.exit(0)

        # If we receive a non-zero system exit update the
        # status to reflect an error. Exit with code 1.
        logging.info('Workflow encountered an error.')
        logging.info('See nextflow output for details.')
        invocation.status = Statuses.ENCOUNTERED_ERROR
        db.commit()
        sys.exit(1)

    # This error is thrown if the path to Nextflow
    # is not available, and therefore cannot be launched
    except FileNotFoundError as e:
        logging.info(f"Cant find '{nextflow}' on the path.")
        logging.info(e)
        invocation.status = Statuses.ENCOUNTERED_ERROR
        db.commit()
        sys.exit(1)

    # Handle all other exception classes in the event of
    # unhandled exceptions occurring within the callable.
    # Set the status to error and exit with code 1.
    except Exception as e:
        logging.info('Workflow encountered an error.')
        logging.info(e)
        invocation.status = Statuses.ENCOUNTERED_ERROR
        db.commit()
        sys.exit(1)


def main():
    """Parse arguments and launch a workflow."""
    args = parse_args()
    invoke(
        id=args.id,
        workflow=args.workflow,
        params=args.params,
        revision=args.revision,
        work_dir=args.work_dir,
        log_file=args.log_file,
        rpc_port=args.rpc_port,
        meta_file=args.meta_file,
        std_out=args.std_out,
        database=args.database,
        nextflow=args.nextflow,
        wsl=args.wsl)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""Combine documentation files and write them out."""

import argparse
from collections import OrderedDict
import json
import os
import sys
from typing import Dict, Sequence, Tuple


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Combine documentation files and write them out.",
        usage=(
            "parse_docs -p docs -s intro usage "
            "-oj nextflow_config.json -ot README.md "
            "-ns nextflow_schema.json -al"
        )
    )

    parser.add_argument(
        '-p',
        '--paths',
        required=True,
        nargs="+",
        help='Locations (files or dirs) in which to find doc files.'
    )

    parser.add_argument(
        '-s',
        '--sections',
        required=True,
        nargs="+",
        help='Names of required documentation sections.'
    )

    parser.add_argument(
        '-e',
        '--extensions',
        nargs="+",
        required=True,
        default=['.rst', '.txt', '.md'],
        help='List of allowed extensions.'
    )

    parser.add_argument(
        '-oj',
        '--output_json',
        help='Path at which to output JSON.'
    )

    parser.add_argument(
        '-ot',
        '--output_text',
        help='Path at which to output text file.'
    )

    parser.add_argument(
        '-ns',
        '--nextflow_schema',
        required=True,
        help='Nextflow schema JSON.'
    )

    parser.add_argument(
        '-od',
        '--output_definition',
        required=False,
        help='Output definition JSON.'
    )

    parser.add_argument(
        '-nc',
        '--nextflow_config',
        required=False,
        default='nextflow.config',
        help='Nextflow config file.'
    )

    parser.add_argument(
        '-hp',
        '--hide_params',
        required=False,
        nargs="+",
        default=[],
        help='Input parameters to be hidden in the input options table.'
    )
    parser.add_argument(
        '-al',
        '--add_docs_link',
        required=False,
        action='store_true',
        help='Append text and a link to the epi2me-docs.'
    )
    args = parser.parse_args(sys.argv[1:])
    return args


def load_file(
    name: str,
    path: str
) -> Dict[str, str]:
    """Load documentation from file."""
    with open(path) as read_in:
        contents = "".join(read_in.readlines())
        return {name: contents}


def check_path(
    path: str,
    required: Sequence[str],
    extensions: Sequence[str]
) -> Tuple[bool, str]:
    """Check if a path exists and has the right extension."""
    basename = os.path.basename(path)
    nameroot, nameext = os.path.splitext(basename)
    if (
        os.path.isfile(path)
        and nameroot in required
        and nameext in extensions
    ):
        return True, nameroot
    return False, nameroot


def load(
    paths: Sequence[str],
    required: Sequence[str],
    extensions: Sequence[str]
) -> Dict[str, str]:
    """Load documentation from paths."""
    sections = {}
    for item in paths:
        check, name = check_path(
            item, required, extensions)
        if check:
            sections.update(load_file(name, item))
            continue
        if os.path.isdir(item):
            for subitem in os.listdir(item):
                subitem_path = f"{item}/{subitem}"
                subcheck, subname = check_path(
                    subitem_path, required, extensions)
                if not subcheck:
                    continue
                sections.update(
                    load_file(subname, subitem_path))
    return sections


def check(
    found: Dict[str, str], required: Sequence[str]
) -> None:
    """Check documentation is in order."""
    notfound_reqs = []
    for req in required:
        if not found.get(req):
            notfound_reqs.append(req)

    if notfound_reqs:
        notfound = ", ".join(notfound_reqs)
        print(
            "Error: Not all expected sections "
            f"were found: {notfound}")
        sys.exit(1)

    return OrderedDict(
        [(req, found[req]) for req in required])


def write_json(
    sections: Dict[str, str], path: str, key: str = 'docs'
) -> None:
    """Write documentation to json."""
    data = {}
    if os.path.exists(path):
        with open(path, "r") as existing_file:
            try:
                data.update(json.load(existing_file))
            except json.decoder.JSONDecodeError:
                raise RuntimeError(
                    "Error: Output JSON file exists but cannot "
                    f"be loaded, it may be corrupt: {path}")
    data.update({key: sections})
    with open(path, "w") as out_file:
        json.dump(data, out_file, indent=4)


def join_sections(
    sections: Dict[str, str], workflow_title
) -> str:
    """Join selected docs sections."""
    title_dic = {
        "01_brief_description": str(workflow_title),
        "02_introduction": "Introduction",
        "03_compute_requirements": "Compute requirements",
        "04_install_and_run": "Install and run",
        "05_input_example": "Input example",
        "06_pipeline_overview": "Pipeline overview",
        "07_input_parameters": "Input parameters",
        "08_outputs": "Outputs",
        "09_related_protocols": "Related protocols",
        "10_troubleshooting": "Troubleshooting",
        "11_FAQ": "FAQs",
        "12_other": "Related blog posts"}
    joined_sections = ''
    for k, v in sections.items():
        title = '## {}\n\n'.format(str(title_dic[k]))
        if k == "01_brief_description":
            title = '# {}\n\n'.format(str(title_dic[k]))
        joined_sections += title
        doc_text = '{}\n\n\n\n'.format(v)
        joined_sections += doc_text
    return joined_sections


def dict_raise_on_duplicates(ordered_pairs):
    """Reject duplicate keys."""
    d = {}
    for k, v in ordered_pairs:
        if k in d:
            raise ValueError(
                "output_definition.json contains duplicate key: {}"
                .format(k))
        else:
            d[k] = v
    return d


def create_outputs(
        path: str,
        docs_folder: str,
) -> None:
    """Create outputs.md from output_definitions.json of workflow."""
    with open(path, 'r') as f:
        d = json.load(f, object_pairs_hook=dict_raise_on_duplicates)
        files = d['files']
    output_md = os.path.join(docs_folder[0], "08_outputs.md")
    with open(output_md, 'w+') as f:
        f.write(
            "Output files may be aggregated including information for all "
            "samples or provided per sample. Per-sample files "
            "will be prefixed with respective aliases and represented "
            "below as {{ alias }}.\n\n"
        )
        f.write(
            "| Title | File path | Description | Per sample or aggregated |\n")
        f.write(
            "|-------|-----------|-------------|--------------------------|\n")
        check_duplicates = list()
        for v in files.values():
            title = v['title']
            if title not in check_duplicates:
                check_duplicates.append(title)
            else:
                raise ValueError(
                    "output_definition.json contains duplicate title: {}"
                    .format(title))
            f.write(
                "| {} | {} | {} | {} |\n".format(
                    title, v["filepath"], v["description"], v["type"]))


def create_compute_requirements(
        path: str,
        docs_folder: str,
) -> None:
    """Create compute_requirements.md from nextflow_schema.json."""
    with open(path) as f:
        d = json.load(f)
        resources = d['resources']
    compute_md = os.path.join(docs_folder[0], "03_compute_requirements.md")
    with open(compute_md, 'w+') as f:
        f.write('Recommended requirements:\n\n')
        f.write('+ CPUs = {}\n'.format(resources["recommended"]["cpus"]))
        f.write('+ Memory = {}\n\n'.format(resources["recommended"]["memory"]))
        f.write('Minimum requirements:\n\n')
        f.write('+ CPUs = {}\n'.format(resources["minimum"]["cpus"]))
        f.write('+ Memory = {}\n\n'.format(resources["minimum"]["memory"]))
        f.write('Approximate run time: {}\n\n'.format(resources["run_time"]))
        f.write(
            'ARM processor support: {}\n'
            .format(str(resources["arm_support"])))


def create_brief_description(
        path: str,
        docs_folder: str,
) -> None:
    """Create brief_description.md from nextflow_schema.json of workflow."""
    with open(path) as f:
        d = json.load(f)
        description = d['description']
    desc_md = os.path.join(docs_folder[0], "01_brief_description.md")
    with open(desc_md, 'w+') as f:
        f.write(description)


def create_install_and_run(
        workflow_title: str,
        slug: str,
        docs_folder: str,
        demo_url: str | None,
        nextflow_config: str,
) -> None:
    """Create 04_intall_and_run.md."""
    run_example = os.path.join(docs_folder[0], "04_install_and_run.md")
    with open(nextflow_config, 'r') as f:
        # parse command from config
        example_cmd = f.read().split('example_cmd = [')[1].split(']')[0]
        # remove any newlines and white space
        example_cmd = example_cmd.replace('"', '') \
            .replace('\n', '').replace('  ', '')
        # Split on commas
        example_cmd_rows = example_cmd.split(',')
        # remove any empty strings
        example_cmd_rows = [x for x in example_cmd_rows if x]
        if '-profile standard' not in example_cmd_rows:
            example_cmd_rows.append('-profile standard')
        # Reformat with \n and \t at commas
        example_cmd = ' \\\n\t'.join(example_cmd_rows)
    with open(run_example, 'w+') as f:
        f.write(f"""
These are instructions to install and run the workflow on command line.
You can also access the workflow via the
[EPI2ME Desktop application](https://epi2me.nanoporetech.com/downloads/).

The workflow uses [Nextflow](https://www.nextflow.io/) to manage
compute and software resources,
therefore Nextflow will need to be
installed before attempting to run the workflow.

The workflow can currently be run using either
[Docker](https://docs.docker.com/get-started/)
or [Singularity](https://docs.sylabs.io/guides/3.0/user-guide/index.html)
to provide isolation of the required software.
Both methods are automated out-of-the-box provided
either Docker or Singularity is installed.
This is controlled by the
[`-profile`](https://www.nextflow.io/docs/latest/config.html#config-profiles)
parameter as exemplified below.

It is not required to clone or download the git repository
in order to run the workflow.
More information on running EPI2ME workflows can
be found in the
[documentation](https://epi2me.nanoporetech.com/epi2me-docs/wfquickstart/).

The following command can be used to obtain the workflow.
This will pull the repository in to the assets folder of
Nextflow and provide a list of all parameters
available for the workflow as well as an example command:

```
nextflow run {workflow_title} --help
```
To update a workflow to the latest version on the command line use
the following command:
```
nextflow pull {workflow_title}
```
""")
        if demo_url is not None:
            f.write(f"""
A demo dataset is provided for testing of the workflow.
It can be downloaded and unpacked using the following commands:
```
wget {demo_url}
tar -xzvf {slug}-demo.tar.gz
```
The workflow can then be run with the downloaded demo data using:
```
nextflow run {workflow_title} \\
\t{example_cmd}
```
""")


def create_inputs(
        path: str,
        docs_folder: str,
        hide_params: list,
) -> None:
    """Create input_parameters.md from nextflow_schema.json of workflow."""
    with open(path) as f:
        d = json.load(f)
        files = d['definitions']
    input_md = os.path.join(docs_folder[0], "07_input_parameters.md")

    with open(input_md, 'w+') as f:
        for _, v in files.items():
            # Count hidden params and don't add the section if all are hidden
            # Usually "hidden" value doesn't exist in the properties
            hidden_values = set()
            for p in v["properties"].keys():
                if 'hidden' in v["properties"][p].keys():
                    if v["properties"][p]["hidden"]:
                        hidden_values.add(p)
                # Additionally hide any params in the hide_params arg
                if p in hide_params:
                    hidden_values.add(p)

            # Don't add form sections that have no parameters
            if len(v["properties"]) == 0:
                continue
            # Don't add form sections if all parameters are hidden
            elif len(v["properties"]) == len(hidden_values):
                continue
            else:
                f.write("### {}\n\n".format(v["title"]))
                f.write(
                    "| Nextflow parameter name  | Type | Description | Help | Default |\n")  # noqa
                f.write("|--------------------------|------|-------------|------|---------|\n")  # noqa
                for prop, values in v["properties"].items():
                    # Don't add hidden parameters to the tables
                    # but check hidden=true
                    if prop in hidden_values:
                        pass
                    else:
                        f.write("| {} | {} | {} | {} | {} |\n".format(
                                prop,
                                values["type"],
                                values["description"],
                                values.get("help_text", ""),
                                values.get("default", "")))
                f.write('\n\n')


def main() -> None:
    """Parse arguments and launch a workflow."""
    args = parse_args()
    link_str = """
For full documentation, including the pipeline overview,
input and output explanations, and FAQs see the
[EPI2ME documentation](https://epi2me.nanoporetech.com/epi2me-docs/workflows/{}).
"""  # noqa: E501
    print(args)
    if args.output_definition:
        create_outputs(args.output_definition, args.paths)
    hide_params = set(["disable_ping"])
    hide_params.update(args.hide_params)
    create_inputs(args.nextflow_schema, args.paths, hide_params)
    create_compute_requirements(args.nextflow_schema, args.paths)
    create_brief_description(args.nextflow_schema, args.paths)
    with open(args.nextflow_schema) as f:
        d = json.load(f)
        workflow_title = d['workflow_title']
        workflow_name = d['title']
        slug = d['title'].replace("epi2me-labs/", "")
        demo = d.get('demo_url')
    create_install_and_run(
        workflow_name, slug, args.paths, demo, args.nextflow_config)
    sections = load(args.paths, args.sections, args.extensions)
    ordered = check(sections, args.sections)
    if args.output_json:
        write_json(ordered, args.output_json)
    if args.output_text:
        output = join_sections(ordered, workflow_title)
        if args.add_docs_link:
            output += link_str.format(slug)
        with open(args.output_text, "w") as new_file:
            new_file.write(output)

    if not (args.output_json or args.output_text):
        print(ordered)


if __name__ == '__main__':
    main()



"""Launch and track nextflow workflows."""

from dataclasses import dataclass
from datetime import datetime
import json
from multiprocessing.connection import Client
import os
import re
import shutil
import socket
import subprocess
import sys
import time
from typing import Callable, Dict, List, Optional, Tuple, TypedDict, Union
from urllib.parse import quote_plus

import psutil
import requests
import shortuuid

from epi2melabs.workflows.database import get_session, Instance, Statuses


class LauncherException(Exception):
    """A catch-all for application exceptions."""

    ...


class PathResponse(TypedDict):
    """Expected fields in path response."""

    name: str
    path: str
    isdir: bool
    exists: bool
    error: Optional[str]
    updated: Optional[float]
    breadcrumbs: Optional[List['PathResponse']]


class FilePathResponse(PathResponse):
    """Expected fields in file response."""

    contents: Union[None, List[str]]


class DirectoryPathResponse(PathResponse):
    """Expected fields in directory response."""

    contents: Union[None, List[PathResponse]]


class InstanceCreateResponse(TypedDict):
    """Expected fields in instance create response."""

    created: bool
    instance: dict
    error: Union[str, None]


@dataclass
class Workflow:
    """Representation of a nextflow workflow."""

    url: str
    name: str
    desc: str
    path: str
    target: str
    schema: Dict
    defaults: Dict
    demo_data: Dict
    docs: Dict

    def to_dict(self):
        """Return workflow data as dict."""
        return {
            'url': self.url,
            'name': self.name,
            'desc': self.desc,
            'path': self.path,
            'target': self.target,
            'schema': self.schema,
            'defaults': self.defaults,
            'demo_data': self.demo_data,
            'docs': self.docs}


class WorkflowLauncher(object):
    """Manages workflow CRUD operations."""

    MAINNF: str = 'main.nf'
    PARAMS: str = 'params.json'
    SCHEMA: str = 'nextflow_schema.json'

    def __init__(
        self, base_dir=None, workflows_dir=None,
        invoker='invoke_nextflow', nextflow='nextflow',
        labsversion='unknown'
    ) -> None:
        """Initialise instance of WorkflowLauncher."""
        self._workflows: Dict[str, Workflow] = {}

        self.curr_dir = os.getcwd()
        self.base_dir = base_dir or os.path.join(
            os.getcwd(), 'nextflow')
        self.workflows_dir = workflows_dir or os.path.join(
            self.base_dir, 'workflows')
        self.instance_dir = os.path.join(
            self.base_dir, 'instances')
        self.database_uri = f'sqlite:///{self.base_dir}/db.sqlite'

        self._invoker = invoker
        self.nextflow = nextflow
        if self.platform in ["win32"]:
            self.nextflow = self._get_wslpath(nextflow)

        self.get_or_create_dir(self.base_dir)
        self.get_or_create_dir(self.instance_dir)
        self.get_or_create_dir(self.workflows_dir)
        self.db = get_session(self.database_uri)

        self.labsversion = labsversion

    @property
    def invoker(self) -> List[str]:
        """Get the command for the nextflow launch script."""
        # Determine if the invoker is a subcommand or not
        if isinstance(self._invoker, list):
            return self._invoker
        return [self._invoker]

    @property
    def platform(self) -> str:
        """Get launcher platform via remote."""
        return sys.platform

    #
    # Workflows
    #
    @property
    def workflows(self) -> Dict:
        """Get data for all visible workflows."""
        for item in os.listdir(self.workflows_dir):
            if self._workflows.get(item):
                continue
            path = os.path.join(self.workflows_dir, item)
            if not os.path.isdir(path):
                continue
            if self.SCHEMA not in os.listdir(path):
                continue
            try:
                self._workflows[item] = self.load_workflow(
                    item, path)
            except LauncherException:
                pass
        return {
            workflow.name: workflow.to_dict()
            for workflow in self._workflows.values()
        }

    def get_workflow(self, name: str) -> Dict:
        """Get data for a single workflow."""
        if workflow := self.workflows.get(name):
            return workflow
        return {}

    def _get_workflow(self, name: str) -> Union[Workflow, None]:
        """Get a workflow instance."""
        if not self.workflows:
            pass
        if workflow := self._workflows.get(name):
            return workflow
        return None

    def load_workflow(self, name: str, path: str) -> Workflow:
        """Create a workflow instance."""
        schema = self._load_schema(os.path.join(path, self.SCHEMA))

        if not schema:
            raise LauncherException(
                f'Cannot reload {name}: missing schema')

        defaults = self._get_schema_defaults(schema)
        demo_data = self._get_schema_demo_data(schema)

        target = None
        main_nf = os.path.join(path, self.MAINNF)
        if os.path.exists(main_nf):
            target = main_nf

        if not target:
            target = schema.get('title', ' ')
            if ' ' in target:
                raise LauncherException(
                    'Cannot load {name}: invalid title')

        if not target:
            raise LauncherException(
                'Cannot load {name}: no entrypoint')

        return Workflow(
            name=name,
            path=path,
            target=target,
            schema=schema,
            defaults=defaults,
            demo_data=demo_data,
            url=schema.get('url', '#'),
            desc=schema.get('description', ''),
            docs=schema.get('docs', {}))

    #
    # Schema
    #
    def _load_schema(self, path: str) -> Dict:
        """Load data from a workflow schema."""
        with open(path, "r") as schema_path:
            try:
                schema = json.load(schema_path)
                self._annotate_parameter_order(schema)
                return schema
            except json.decoder.JSONDecodeError:
                return {}

    def _annotate_parameter_order(self, schema: Dict) -> Dict:
        """Annotate parameters in a workflow schema with order field."""
        sections = schema.get("definitions", {})
        for _, section in sections.items():
            for index, param in enumerate(
                section.get(
                    "properties", {}
                ).values()
            ):
                param['order'] = index
        return sections

    def _get_schema_demo_data(
        self, schema: Dict
    ) -> Dict[str,  Union[str, int, float, bool]]:
        """Extract demo_data key values from parameters in a schema."""
        return self._get_param_key_value_pairs_from_schema(
            schema, self._get_parameter_demo_data)

    def _get_parameter_demo_data(
        self, param: Dict
    ) -> Union[str, int, float, bool, None]:
        """Extract demo_data value for a parameter in a schema."""
        return param.get('demo_data')

    def _get_schema_defaults(
        self, schema: Dict
    ) -> Dict[str,  Union[str, int, float, bool]]:
        """Extract default key values from parameters in a schema."""
        return self._get_param_key_value_pairs_from_schema(
            schema, self._get_parameter_default)

    def _get_parameter_default(
        self, param: Dict
    ) -> Union[str, int, float, bool, None]:
        """Extract default value for a parameter in a schema."""
        default = None
        if 'default' in param:
            default = self._sanitise_parameter_default(
                param['default'], param.get('type'))
        return default

    def _get_param_key_value_pairs_from_schema(
        self, schema: Dict, getter: Callable
    ) -> Dict[str,  Union[str, int, float, bool]]:
        """Apply getter to parameters in a schema and return results."""
        extracted = self._get_param_key_value_pairs_from_schema_section(
            schema.get("properties", {}), getter)
        sections = schema.get("definitions", {})
        for _, section in sections.items():
            extracted = {
                **extracted,
                **self._get_param_key_value_pairs_from_schema_section(
                    section.get("properties", {}), getter)}
        return extracted

    def _get_param_key_value_pairs_from_schema_section(
        self, schema_section: Dict, getter: Callable
    ) -> Dict[str,  Union[str, int, float, bool]]:
        """Apply getter to parameters in a schema section."""
        extracted = {}
        for key, param in schema_section.items():
            default = getter(param)
            if default is None:
                continue
            extracted[key] = default
        return extracted

    def _sanitise_parameter_default(
        self,
        default: Union[str, int, float, bool],
        _type=None
    ) -> Union[str, int, float, bool]:
        """Coerce schema defaults to appropriate pythom types."""
        if not _type:
            return default
        elif _type == "boolean":
            if not isinstance(default, bool):
                default = (default == "true")
            return default
        elif isinstance(default, str) and default.strip() == "":
            return ""
        elif _type == "integer":
            return int(default)
        elif _type == "number":
            return float(default)
        return str(default)

    #
    # Instances
    #
    @property
    def instances(self) -> Dict:
        """Get data for all workflow run instances."""
        return {
            instance.id: instance.to_dict()
            for instance in self.db.query(Instance).all()
        }

    def get_instance(self, id: str) -> Dict:
        """Get data for a single workflow run instance."""
        if instance := self._get_instance(id):
            return instance.to_dict()
        return {}

    def _get_instance(self, id: str) -> Union[Instance, None]:
        """Get a workflow run instance."""
        if instance := self.db.query(Instance).get(id):
            return instance
        return None

    def create_instance(
        self, name: str, workflow_name: str, params: Dict
    ) -> InstanceCreateResponse:
        """Create a new workflow run instance if workflow is available."""
        workflow = self._get_workflow(workflow_name)

        if not workflow or not workflow.target:
            return InstanceCreateResponse(
                created=False, instance={},
                error='Could not create instance, workflow unknown')

        return self._create_instance(
            name, workflow.name, workflow.target,
            workflow.schema, params)

    def _create_instance(
        self, name: str, workflow_name: str, workflow_target: str,
        workflow_schema: Dict, params: Dict
    ) -> InstanceCreateResponse:
        """Create and launch new workflow run instance."""
        # Generate instance details
        _id = str(shortuuid.uuid())
        now = datetime.now().strftime("%Y-%m-%d-%H-%M")
        dirname = '_'.join([now, workflow_name, _id])
        path = os.path.join(self.instance_dir, dirname)

        # Create instance db record
        instance = Instance(_id, path, name, workflow_name)
        self.db.add(instance)
        self.db.commit()

        # Construct instance filepaths
        params_file = os.path.join(path, self.PARAMS)
        nf_logfile = os.path.join(path, 'nextflow.log')
        nf_std_out = os.path.join(path, 'nextflow.stdout')
        iv_std_out = os.path.join(path, 'invoke.stdout')
        meta_file = os.path.join(path, 'meta.json')
        out_dir = os.path.join(path, 'output')
        work_dir = os.path.join(path, 'work')

        # Touch instance files and directories
        self.get_or_create_dir(path)
        self.get_or_create_dir(out_dir)
        self.get_or_create_dir(work_dir)

        for targ in [nf_logfile, nf_std_out, iv_std_out]:
            with open(targ, 'a'):
                pass

        # Coerce params and write to file
        params = self._fix_parameters(workflow_schema, params)
        # Todo: generalise to support 3rd party wfs
        params['out_dir'] = out_dir
        params['wf'] = {'agent': f'labslauncher/{self.labsversion}'}
        self.write_json(params, params_file)

        # Get version
        wfversion = params['wfversion']
        revision = wfversion if not os.path.exists(workflow_target) else None

        # Get a port for comms with the workflow wrapper
        rpc_port = self.get_free_port()

        # Handle being on windows
        windows = False
        if self.platform in ["win32"]:
            windows = True

            # Make sure nextflow is executable
            chmodx = ['wsl', 'chmod', '+x', self.nextflow]
            proc = popen(chmodx, windows=windows)
            proc.wait()

            # Re-write paths within params file
            wsl_params = self._get_wslpath_params(workflow_schema, params)
            params_file = os.path.join(path, f"wsl-{self.PARAMS}")
            self.write_json(wsl_params, params_file)
            params_file = self._get_wslpath(params_file)

            # Re-write nextflow command paths
            work_dir = self._get_wslpath(work_dir)
            nf_logfile = self._get_wslpath(nf_logfile)
            if os.path.exists(workflow_target):
                workflow_target = self._get_wslpath(workflow_target)

        # Launch process
        error = self._start_instance(
            instance, workflow_target, params_file, work_dir, nf_logfile,
            nf_std_out, iv_std_out, meta_file, rpc_port, self.database_uri,
            revision, windows)

        return InstanceCreateResponse(
            created=True, instance=instance.to_dict(), error=error)

    def _start_instance(
        self, instance: Instance, workflow_target: str, params_file: str,
        work_dir: str, nf_logfile: str, nf_std_out: str, iv_std_out: str,
        meta_file: str, rpc_port: str, database: str,
        revision: Union[str, None], windows: bool
    ) -> Union[str, None]:
        """Launch new workflow run instance."""
        command = self.invoker + [
            '-i', instance.id, '-w', workflow_target, '-m', meta_file,
            '-p', params_file, '-wd', work_dir, '-l', nf_logfile,
            '-s', nf_std_out, '-d', database, '-n', self.nextflow,
            '-rpc', rpc_port]

        if windows:
            command = command + ['-wsl']

        if revision:
            command = command + ['-r', revision]

        # Ensure the default location for docker on MacOS is available
        env = os.environ.copy()
        env["PATH"] = "/usr/local/bin:" + env["PATH"]

        # Setup process fds
        logfile = open(iv_std_out, 'a')
        stdout = logfile
        stderr = logfile

        # Create settings map
        kwargs = dict(
            stdout=stdout, stderr=stderr, close_fds=True,
            cwd=self.base_dir, env=env)

        # Update settings per platform
        if windows:
            newgrp = subprocess.CREATE_NEW_PROCESS_GROUP  # type: ignore
            kwargs.update(creationflags=newgrp)
        else:
            kwargs.update(start_new_session=True)

        # Launch invocation script
        try:
            proc = popen(command, windows=windows, **kwargs)
        except FileNotFoundError:
            error = 'Could not launch instance, invocation script not found'
            logfile.write(error)
            instance.status = Statuses.ENCOUNTERED_ERROR  # type: ignore
            self.db.commit()
            return error

        instance.pid = proc.pid  # type: ignore
        self.db.commit()

    def delete_instance(self, id: str, delete: bool = False) -> bool:
        """Delete a workflow run instance."""
        instance = self._get_instance(id)
        if not instance:
            return False

        # Stop any process
        actioned = self._stop_instance(instance)
        if actioned:
            instance.status = Statuses.TERMINATED  # type: ignore
            self.db.commit()

        # Optionally delete the directory
        if delete:
            try:
                shutil.rmtree(instance.path)
            except FileNotFoundError:
                pass

            # Delete record
            self.db.delete(instance)
            self.db.commit()

        return True

    def _stop_instance(self, instance) -> bool:
        """Stop a workflow run instance from running."""
        if instance.status != Statuses.LAUNCHED:
            return False

        meta_file = os.path.join(instance.path, 'meta.json')
        meta_data = {}

        for _ in range(3):
            try:
                with open(meta_file, 'r', encoding='utf8') as mf:
                    meta_data = json.load(mf)
                    break
            except Exception:
                time.sleep(2)

        if rpc_port := meta_data.get('rpc_port', None):
            address = ('localhost', int(rpc_port))
            conn = Client(address, authkey=bytes(
                instance.id, encoding='utf8'))
            conn.send(('close', instance.id))
            conn.close()
            return True
        return False

    #
    # Pathing
    #
    def get_path(self, path: str) -> PathResponse:
        """Get information on a filesystem path."""
        # Explicitly handle special 'root' keyword
        if path.lower() == 'root':
            return self._get_root_path_response()
        path = self._fix_path(path)
        return self._get_path_response(path)

    def get_file(
        self,
        path: str,
        contents: bool = False
    ) -> FilePathResponse:
        """Get information on a filesystem file path."""
        lines = None
        path = self._fix_path(path)
        resp = self._get_path_response(path)
        if resp.get('exists'):
            if contents:
                lines = self.read_file(path)
            if resp.get('isdir'):
                resp['error'] = 'Path is not a file'
        return FilePathResponse(**resp, contents=lines)

    def read_file(self, path: str) -> List[str]:
        """Load file data."""
        lines = []
        with open(path) as lf:
            lines = lf.readlines()
            lines = [self._process_line(line) for line in lines]
        return lines

    def _process_line(self, line: str) -> str:
        """Strip a line of text of ANSI sequences."""
        line = line.rstrip()
        # 7-bit C1 ANSI sequences
        # https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
        ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
        return ansi_escape.sub('', line)

    def get_directory(
        self,
        path: str,
        contents: bool = False
    ) -> DirectoryPathResponse:
        """Get information on a filesystem directory path."""
        # Explicitly handle special 'root' keyword
        if path.lower() == 'root':
            return self._get_root_directory_response(
                self.platform)
        items = None
        path = self._fix_path(path)
        resp = self._get_path_response(path)
        if resp.get('exists'):
            if contents:
                items = self.list_dir(path)
            if not resp.get('isdir'):
                resp['error'] = 'Path is not a directory'
        return DirectoryPathResponse(**resp, contents=items)

    def list_dir(self, path: str) -> List[PathResponse]:
        """List the contents of a directory."""
        items = []
        for item in os.listdir(path):
            item_abspath = self._fix_path(
                os.path.join(path, item))
            items.append(self._get_path_response(item_abspath))
        return items

    def _get_root_path_response(
        self,
    ) -> PathResponse:
        resp = PathResponse(
            name='root', path='root', updated=None,
            exists=True, error=None, isdir=True, breadcrumbs=None)
        return resp

    def _get_root_directory_response(
        self,
        platform: str,
    ) -> DirectoryPathResponse:
        resp = self._get_root_path_response()
        if platform in ["win32"]:
            items = self._get_drives()
            return DirectoryPathResponse(**resp, contents=items)
        items = self.list_dir('/')
        return DirectoryPathResponse(**resp, contents=items)

    def _get_drives(self):
        """Get the contents of 'root' (drives) on Windows."""
        partitions = psutil.disk_partitions()
        return [self._get_path_response(
            d.device.replace(os.sep, '/')) for d in partitions]

    def _get_path_response(
        self,
        abspath: str,
        breadcrumbs=True,
        exists_error="Path does not exist or host cannot see it."
    ) -> PathResponse:
        """Generate an undifferentiated path response."""
        name = os.path.basename(abspath) or abspath
        exists = os.path.exists(abspath)
        is_dir = os.path.isdir(abspath)
        error = None if exists else exists_error
        modified = os.path.getmtime(abspath) if exists else None
        breadcrumbs = self._get_breadcrumbs(
            abspath) if breadcrumbs else None
        return PathResponse(
            name=name,
            path=abspath,
            updated=modified,
            exists=exists,
            error=error,
            isdir=is_dir,
            breadcrumbs=breadcrumbs)

    def _get_breadcrumbs(
        self,
        path: str,
        _breadcrumbs: Optional[List[PathResponse]] = None
    ) -> List[PathResponse]:
        """Get seperate path responses for each path section."""
        if not os.path.abspath(path):
            raise ValueError('Path is not absolute')
        if not _breadcrumbs:
            _breadcrumbs = []
        head, _ = os.path.split(path)
        if not head or head in ['/', path]:
            return _breadcrumbs
        _breadcrumbs.append(
            PathResponse(
                name=os.path.basename(head) or head,
                path=head,
                updated=None,
                exists=True,
                error=None,
                isdir=True,
                breadcrumbs=None))
        _breadcrumbs = self._get_breadcrumbs(head, _breadcrumbs)
        return _breadcrumbs

    def _fix_path(self, path: str) -> str:
        """Convert a relative filesystem path to absolute."""
        if not os.path.isabs(path):
            path = os.path.join(self.base_dir, path)
        return os.path.abspath(path).replace(os.sep, "/")

    def get_or_create_dir(self, path: str) -> Tuple[bool, str]:
        """Get or create a directory."""
        if not os.path.exists(path):
            os.mkdir(path)
        if os.path.isdir(path):
            return True, os.path.abspath(path)
        return False, os.path.abspath(path)

    def _fix_parameters(
        self, schema: Dict, params: Dict
    ) -> Dict:
        """Convert path-like parameter values to absolute paths."""
        return self._fix_pathing_params(
            schema, params, self._fix_path)

    def _get_wslpath_params(
        self, schema: Dict, params: Dict
    ) -> Dict:
        """Convert path-like parameter values to WSL paths."""
        return self._fix_pathing_params(
            schema, params, self._get_wslpath)

    def _fix_pathing_params(
        self, schema: Dict, params: Dict, fixer: Callable
    ) -> Dict:
        """Apply fixer to each path-like parameter subschema in params."""
        updated = {}
        for name, value in params.items():
            param_schema = self._get_schema_for_param(schema, name)
            fmt = param_schema.get('format')
            if fmt in ['path', 'file-path', 'directory-path']:
                if any(value.startswith(i) for i in
                        ['http', 's3', r'${projectDir}']):
                    updated[name] = value
                    continue
                updated[name] = fixer(value)
                continue
            updated[name] = value
        return updated

    def _get_schema_for_param(
        self, schema: dict, param_name: str
    ) -> Dict:
        """Get the schema section for a given parameter."""
        sections = schema.get('definitions', {})
        for section in sections.values():
            for k, v in section.get('properties', {}).items():
                if k == param_name:
                    return v
        return {}

    def _get_wslpath(self, path: str) -> str:
        """Get the WSL path for a given host path."""
        path = path.replace('\\', '/')
        proc = popen(
            ["wsl", "wslpath", "-a", path],
            windows=True,
            text=True,
            stdout=subprocess.PIPE)
        return proc.stdout.readline().rstrip()

    def write_json(self, data, path):
        """Write data to a json file."""
        with open(path, 'w') as json_file:
            json_file.write(json.dumps(data, indent=4))

    def get_free_port(self):
        """Find a free port."""
        sock = socket.socket()
        sock.bind(('', 0))
        return str(sock.getsockname()[1])


class RemoteWorkflowLauncher(WorkflowLauncher):
    """Manages workflow CRUD operations via remote."""

    def __init__(
        self, base_dir, workflows_dir,
        ip: str = '0.0.0.0', port: str = '8090'
    ) -> None:
        """Initialise instance of RemoteWorkflowLauncher."""
        self.ip = ip
        self.port = port

        super().__init__(base_dir, workflows_dir)

    @property
    def platform(self):
        """Get launcher platform via remote."""
        response = requests.get(
            f'http://{self.ip}:{self.port}/platform')
        data = response.json()
        return data.get('platform', None)

    @property
    def instances(self) -> Dict:
        """Get data for all workflow run instances via remote."""
        response = requests.get(
            f'http://{self.ip}:{self.port}'
            f'/instances?uuid={shortuuid.uuid()}')
        return response.json()

    def get_instance(self, id: str) -> Dict:
        """Get data for a single workflow run instance via remote."""
        response = requests.get(
            f'http://{self.ip}:{self.port}'
            f'/instances/{id}?uuid={shortuuid.uuid()}')
        return response.json()

    def create_instance(
        self, name: str, workflow_name: str, params: Dict
    ) -> InstanceCreateResponse:
        """Create a new workflow run instance via remote."""
        workflow = self._get_workflow(workflow_name)

        if not workflow or not workflow.target:
            return InstanceCreateResponse(
                created=False, instance={},
                error='Could not create instance, workflow unknown')

        response = requests.post(
            f'http://{self.ip}:{self.port}/instances',
            data=json.dumps({
                'name': name,
                'workflow_name': workflow_name,
                'workflow_target': workflow.target,
                'workflow_schema': workflow.schema,
                'params': params,
            }),
            headers={
                'Content-type': 'application/json',
                'Accept': 'text/plain'
            })

        data = response.json()
        return data

    def delete_instance(
        self, id: str, delete: bool = False
    ) -> bool:
        """Delete a workflow run instance via remote."""
        response = requests.delete(
            f'http://{self.ip}:{self.port}'
            f'/instances/{id}?delete={delete}')

        data = response.json()
        return data['deleted']

    def get_path(self, path: str) -> PathResponse:
        """Get information on a filesystem path via remote."""
        response = requests.get(
            f'http://{self.ip}:{self.port}'
            f'/path/{quote_plus(quote_plus(path))}')
        return response.json()

    def get_file(
        self, path: str, contents: bool = False
    ) -> FilePathResponse:
        """Get information on a filesystem file path via remote."""
        _cont = ''
        if contents:
            _cont = f'?contents={shortuuid.uuid()}'
        response = requests.get(
            f'http://{self.ip}:{self.port}'
            f'/file/{quote_plus(quote_plus(path))}{_cont}')
        return response.json()

    def get_directory(
        self, path: str, contents: bool = False
    ) -> DirectoryPathResponse:
        """Get information on a filesystem directory path via remote."""
        _cont = ''
        if contents:
            _cont = f'?contents={shortuuid.uuid()}'
        response = requests.get(
            f'http://{self.ip}:{self.port}'
            f'/directory/{quote_plus(quote_plus(path))}{_cont}')
        return response.json()

    def _get_wslpath(self, path: str) -> str:
        """Get the WSL path for a given host path via remote."""
        response = requests.get(
            f'http://{self.ip}:{self.port}'
            f'/wslpath/{quote_plus(quote_plus(path))}')
        data = response.json()
        return data.get('path', '')


def get_workflow_launcher(
    base_dir, workflows_dir, remote=False,
    ip=None, port=None
):
    """Create the correct workflow launcher."""
    if remote:
        kwargs = {'ip': ip, 'port': port}
        for k, v in kwargs.items():
            if v is None:
                kwargs.pop(k)
        return RemoteWorkflowLauncher(base_dir, workflows_dir, **kwargs)
    return WorkflowLauncher(base_dir, workflows_dir)


def popen(
    cmd: List[str], windows: bool = False, **kwargs
) -> subprocess.Popen:
    """Run subprocess, ensuring no popups on windows."""
    if windows:
        startupinfo = subprocess.STARTUPINFO()  # type: ignore
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW  # type: ignore
        kwargs.update(startupinfo=startupinfo)
    return subprocess.Popen(cmd, **kwargs)
